"""Library and command-line tool for getting stats on various typeshed packages."""

__all__ = [
    "__about__",
    "__author__",
    "__email__",
    "__license__",
    "__summary__",
    "__title__",
    "__version__",
]

# This file is created at build time by hatch-vcs
from ._version import __version__

__title__ = "typeshed-stats"
__summary__ = __about__ = __doc__
__author__ = "Alex Waygood"
__email__ = "alex.waygood@gmail.com"
__license__ = "MIT"
